#include "ParticleSystem.h"


WeakReferenceCountedPointer<Shader> ParticleSystem::s_shader;


ParticleSystem::ParticleSystem(const Texture::Ref& texture, float newParticleLifetime) : 
    m_newParticleLifetime(newParticleLifetime),
    m_texture(texture) {

    // Fetch the cached shader
    m_shader = s_shader.createStrongPtr();
    if (m_shader.isNull()) {
        // Update the cached shader
        s_shader = m_shader = Shader::fromFiles("ParticleSystem.vrt", "ParticleSystem.pix");
    }

    // Allocate the vertex arrays to maximum size
    const int numVertices = MAX_PARTICLES * 4;
    VertexBuffer::Ref buffer =
        VertexBuffer::create((sizeof(Vector3) + sizeof(Vector2)) * numVertices, 
                             VertexBuffer::WRITE_EVERY_FRAME);

    // The texture coordinates repeat around each quad and are always
    // the same, so upload them once at program start.
    Array<Vector2> texCoord;
    texCoord.resize(numVertices);
    for (int p = 0; p < MAX_PARTICLES; ++p) {
        texCoord[p * 4 + 0] = Vector2(1, 0); 
        texCoord[p * 4 + 1] = Vector2(1, 1); 
        texCoord[p * 4 + 2] = Vector2(0, 1); 
        texCoord[p * 4 + 3] = Vector2(0, 0); 
    }

    m_gpuTexCoord = VertexRange(texCoord, buffer);

    // Reserve memory for the vertex array
    Array<Vector3> vertex;
    vertex.resize(numVertices);
    m_gpuPosition = VertexRange(vertex, buffer);
}


void ParticleSystem::spawnParticle(const Vector3& position) {
    if (m_particleArray.size() == MAX_PARTICLES) {
        // Remove some other particle to make room; the one
        // at the front of the array is the farthest from the
        // camera and unlikely to be visible.
        m_particleArray.fastRemove(0);
    }

    Particle& particle = m_particleArray.next();
 
    particle.position   = position;
    particle.velocity   = Vector3::random(m_rnd) * 0.2f + Vector3::unitY();
    particle.life       = m_newParticleLifetime;
    particle.rotation   = m_rnd.uniform(0.0f, 2.0f * pif());
}


void ParticleSystem::simulate() {
    // Fix the simulation speed, regardless of the actual rendering rate
    static const float dt = 1.0f / 60.0f;
    for (int p = 0; p < m_particleArray.size(); ++p) {
        Particle& particle = m_particleArray[p];

        particle.life -= dt;
        if (particle.life <= 0) {
            // Die
            m_particleArray.fastRemove(p);

            // A new element just moved into this position, so we have
            // to process it the next time around the loop
            --p;
        } else {
            particle.position += particle.velocity * dt;
            
            // Random acceleration
            particle.velocity += Vector3::random(m_rnd) * 0.1f * dt;
        }
    }
}


void ParticleSystem::sort(const Vector3& viewVector) {
    // Sort in depth from the camera
    for (int p = 0; p < m_particleArray.size(); ++p) {
        Particle& particle = m_particleArray[p];
        particle.sortKey = viewVector.dot(particle.position);
    }

    m_particleArray.sort(SORT_DECREASING);
}


void ParticleSystem::render
(RenderDevice*         rd,
 const Texture::Ref&   depthBuffer, 
 const GCamera&        camera, 
 const GLight&         light,
 const ShadowMap::Ref& shadowMap) {

    const CFrame& cameraFrame = camera.coordinateFrame();

    // Sort in depth
    sort(cameraFrame.lookVector());

#define FANCY_RENDERING
#ifdef FANCY_RENDERING
    // Update the data array

    // Map the GPU memory into the CPU's address space
    Vector3* vertexPtr = (Vector3*)m_gpuPosition.mapBuffer(GL_WRITE_ONLY);

    // Write directly to the mapped memory
    for (int p = 0; p < m_particleArray.size(); ++p) {
        const Particle& particle = m_particleArray[p];
        // Upload four copies of each vertex; we will convert it to
        // the correct location in the vertex shader using the texture
        // coordinate.  A valid alternative would be using the
        // geometry shader to expand each point into a quad, but that
        // is substantially more complex and we haven't studied the
        // geometry shader yet.
        for (int i = 0; i < 4; ++i) {
            *vertexPtr = particle.position;
            ++vertexPtr;
        }
    }

    m_gpuPosition.unmapBuffer();

    rd->pushState();
    {
        rd->setObjectToWorldMatrix(CFrame());

        // Particles should not overwrite depth
        rd->setDepthWrite(false);
        rd->setBlendFunc(RenderDevice::BLEND_SRC_ALPHA, RenderDevice::BLEND_ONE_MINUS_SRC_ALPHA);

#error "TODO: Set the arguments on the shader.  You need to pass at the very least arguments for `up' and `right' vectors to define the world-space axes along which the individual particle billboard squares will align themselves so that they are always facing the camera.  Up should always match the camera's notion of up, and right should always match the camera's notion of `right'."

        //m_shader->args.set("up",      ??);
        //m_shader->args.set("right",   ??);


        rd->setShader(m_shader);
        rd->beginIndexedPrimitives();
        {
            rd->setVertexArray(m_gpuPosition);
            rd->setTexCoordArray(0, m_gpuTexCoord);
            rd->sendSequentialIndices(PrimitiveType::QUADS, m_particleArray.size() * 4);
        }
        rd->endIndexedPrimitives();
    }
    rd->popState();

#else
    // Code for debugging particle positions independent of GPU rendering
    for (int p = 0; p < m_particleArray.size(); ++p) {
        const Particle& particle = m_particleArray[p];
        Draw::sphere(Sphere(particle.position, particle.radius), rd);
    }
#endif
}

