#ifndef ParticleSystem_h
#define ParticleSystem_h

#include <G3D/G3DAll.h>

/** \brief */
class ParticleSystem : public ReferenceCountedObject {
public:

    typedef ReferenceCountedPointer<ParticleSystem> Ref;

private:
    /** Shared instance; "Weak" means that if this is the last
        instance to the shader, it can be deallocated anyway. 
        Regular ReferenceCountedPointers are strong and make
        objects persist. */
    static WeakReferenceCountedPointer<Shader> s_shader;

    /** We preallocate the GPU vertex arrays to avoid the complexity
        of dynamically managing a buffer, but that limits us to a
        fixed maximum number of particles. */
    enum {MAX_PARTICLES = 2048};

    class Particle {
    public:
        Vector3         position;
        Vector3         velocity;

        /** Counts down to zero. */
        float           life;

        float           rotation;

        /** Distance along the view vector, used for sorting particles
            from front to back during rendering.*/
        float           sortKey;

        /** Used by ParticleSystem::sort */
        bool operator>(const Particle& other) const {
            return sortKey > other.sortKey;
        }

        /** Used by ParticleSystem::sort */
        bool operator<(const Particle& other) const {
            return sortKey < other.sortKey;
        }
    };

    Random              m_rnd;

    float               m_newParticleLifetime;

    Array<Particle>     m_particleArray;

    /** Only the first m_particleArray.size() elements are used. */
    VertexRange         m_gpuPosition;

    VertexRange         m_gpuTexCoord;

    Texture::Ref        m_texture;

    Shader::Ref         m_shader;

    /** Called from render() */
    void sort(const Vector3& viewVector);

    ParticleSystem(const Texture::Ref& texture, float newParticleLifetime);

public:

    /** Create the particle system instance, with no particles in it. 

        \param newParticleLifetime How long particles persist for, in
        seconds.
     */
    static Ref create(const Texture::Ref& texture, float newParticleLifetime) {
        return new ParticleSystem(texture, newParticleLifetime);
    }

    /** Create a new particle at this location. */
    virtual void spawnParticle(const Vector3& position);

    virtual void simulate();
    
    virtual void render(RenderDevice* rd, const Texture::Ref& depthBuffer, const GCamera& camera, const GLight& light, const ShadowMap::Ref& shadowMap);
};

#endif
